"""Finds possible packages to install for executables."""
import os
import sys
import glob
import bisect


MAPFILES = {}
PATH_SEP = ";" if os.name == "nt" else ":"
DEFAULT_CONDA_SUGGEST_PATH = os.environ.get(
    "CONDA_SUGGEST_PATH",
    PATH_SEP.join(
        [
            os.path.join(os.path.expanduser("~"), ".local", "share", "conda-suggest"),
            os.path.join(sys.exec_prefix, "share", "conda-suggest"),
        ]
    ),
)


def _make_finder_func(exe):
    import re

    finder_re = re.compile(f".*{exe}.*")
    func = lambda entry: finder_re.match(entry[0]) is not None
    return func


class Mapfile:
    """Provides a lazy-loading view of the map file as bisectable sequence."""

    def __init__(self, filename):
        self.filename = filename
        self.channel, _, rest = os.path.basename(filename).partition(".")
        self.subdir = rest[:-4]  # strips off '.map'
        # make entries list of (exe, pkg) tuples
        # This turns out to be the fastest way to create this tuple for
        # map files in our size range, runs in ~3.5 ms
        with open(filename) as f:
            s = f.read()
        self.entries = [line.partition(":")[::2] for line in s.splitlines()]

    def exact_find(self, exe):
        """Finds the set of packages for an executable"""
        # a space is less than all other real characters
        left = bisect.bisect_left(self.entries, (exe, " "))
        if left == len(self.entries) or self.entries[left][0] != exe:
            # Executable not found
            if (
                self.subdir.startswith("win")
                and not exe.endswith(".exe")
                and not exe.endswith(".bat")
            ):
                # also search extensions
                return self.exact_find(exe + ".exe") + self.exact_find(exe + ".bat")
            else:
                return []
        # a tilde is greater than other real characters
        right = bisect.bisect_right(self.entries, (exe, "~"), lo=left)
        return self.entries[left:right]

    def substring_find(self, exe, finder_func=None):
        """Finds the package names where the exe is contained in the actual executable name."""
        if finder_func is None:
            finder_func = _make_finder_func(exe)
        return list(filter(finder_func, self.entries))


def get_mapfilenames(conda_suggest_path=None):
    """Get's the mapfilenames on the system."""
    if conda_suggest_path is None:
        conda_suggest_path = DEFAULT_CONDA_SUGGEST_PATH
    mapfilenames = []
    for d in conda_suggest_path.split(PATH_SEP):
        mapfilenames.extend(glob.iglob(os.path.join(d, "*.map")))
    return mapfilenames


def get_mapfile(filename):
    """Gets a mapfile from the cache or makes a new one"""
    global MAPFILES
    mapfile = MAPFILES.get(filename, None)
    if mapfile is None:
        mapfile = MAPFILES[filename] = Mapfile(filename)
    return mapfile


def exact_find_in_mapfile(exe, filename):
    """Finds a list of (channel, subdir, executable, package) tuples in a mapfile from the executable."""
    mapfile = get_mapfile(filename)
    return {(mapfile.channel, mapfile.subdir, e, p) for e, p in mapfile.exact_find(exe)}


def exact_find(exe, conda_suggest_path=None):
    """Finds the package names when the exe exactly matches what is in the package"""
    filenames = get_mapfilenames(conda_suggest_path=conda_suggest_path)
    finds = set()
    for filename in filenames:
        finds |= exact_find_in_mapfile(exe, filename)
    return finds


def substring_find_in_mapfile(exe, filename, finder_func=None):
    """Finds a list of (channel, subdir, executable, package) tuples in a mapfile from the executable substring."""
    mapfile = get_mapfile(filename)
    return {
        (mapfile.channel, mapfile.subdir, e, p)
        for e, p in mapfile.substring_find(exe, finder_func=finder_func)
    }


def substring_find(exe, conda_suggest_path=None, finder_func=None):
    """Finds the package names where the exe is contained in the actual executable name."""
    filenames = get_mapfilenames(conda_suggest_path=conda_suggest_path)
    if finder_func is None:
        finder_func = _make_finder_func(exe)
    finds = set()
    for filename in filenames:
        finds |= substring_find_in_mapfile(exe, filename, finder_func=finder_func)
    return finds


def message_string(exe, conda_suggest_path=None):
    """Produces a 'command-not-found' message, suitable for printing."""
    exact = exact_find(exe, conda_suggest_path=conda_suggest_path)
    if exact:
        s = f"Command {exe!r} not found in the environment, but can be installed with any of:\n\n"
        uniques = sorted({(channel, pkg) for channel, _, _, pkg in exact})
        for channel, pkg in uniques:
            s += f"    $ conda install -c {channel} {pkg}\n"
        return s[:-1]
    substrings = substring_find(exe, conda_suggest_path=conda_suggest_path)
    if substrings:
        s = f"Command {exe!r} not found in a conda package, perhaps you meant:\n\n"
        uniques = sorted({(channel, cmd, pkg) for channel, _, cmd, pkg in substrings})
        for channel, cmd, pkg in uniques:
            s += f"    {cmd} (conda install -c {channel} {pkg})\n"
        return s[:-1]
    s = f"Command {exe!r} not found in any conda package."
    return s


def message(exe, conda_suggest_path=None):
    """Prints a 'command-not-found' message."""
    s = message_string(exe, conda_suggest_path=conda_suggest_path)
    print(s)



"""Generates the cache files (which should not be distributed) and the map files
(which should).
"""
import os
import re
import json
import tarfile

import tqdm


SUBDIRS = [
    "noarch",
    "linux-64",
    "osx-64",
    # "osx-arm64",
    "win-64",
    "linux-ppc64le",
    "linux-aarch64",
]
EXECUTABLE_RE = re.compile(b"(?:bin|Scripts)/([^/]*)")
DEFAULT_REMOVE_EXPRS = ("__pycache__",)


def _get_repodata_packages(channel, subdir):
    repodata_json = f"{channel}/{subdir}/repodata.json"
    print(f"Loading {repodata_json}")
    with open(repodata_json) as f:
        repodata = json.load(f)
    return repodata["packages"]


def _save_cache(cache, cachefile, display=True):
    if display:
        print(f"Saving cache to {cachefile}")
    with open(cachefile, "w") as f:
        json.dump(cache, f, sort_keys=True, indent=1)


def _add_info_entrypoints(executables, tf):
    try:
        fb = tf.extractfile("info/link.json")
    except KeyError:
        # no link.json in the file
        return
    link = json.load(fb)
    noarch = link.get("noarch", None)
    if noarch is None:
        return
    points = noarch.get("entry_points", ())
    for point in points:
        exe, _, _ = point.partition("=")
        executables.add(exe.strip())


def _add_artifact_to_cache(cache, pkg, channel, subdir, artifact):
    url = f"{channel}/{subdir}/{artifact}"
    name = pkg["name"]
    entry = {
        "package": name,
        "executables": [],
    }
    if not artifact.endswith(".tar.bz2"):
        # skip non-bz2 artifacts for now
        cache[artifact] = entry
        return
    executables = set()
    with tarfile.open(url, mode="r:bz2") as tf:
        # first read files
        fb = tf.extractfile("info/files")
        files = fb.read().splitlines()
        for f in files:
            m = EXECUTABLE_RE.match(f)
            if m is None:
                continue
            executables.add(m.group(1).decode())
        # next add entrypoints
        _add_info_entrypoints(executables, tf)
    entry["executables"] = sorted(executables)
    cache[artifact] = entry


def _get_channel_name(channel):
    return channel.rpartition("/")[2]


def make_cache(channel, subdir):
    """Reads and/or generates the cachefile and returns the cache"""
    # load cache
    channel_name = _get_channel_name(channel)
    cachefile = f"{channel_name}.{subdir}.cache.json"
    if os.path.exists(cachefile):
        print(f"Loading cache from {cachefile}")
        with open(cachefile) as f:
            cache = json.load(f)
    else:
        cache = {}
    # load repodata
    pkgs = _get_repodata_packages(channel, subdir)
    # add packages to cache
    needed = set(pkgs.keys()) - set(cache.keys())
    for i, artifact in enumerate(tqdm.tqdm(needed)):
        _add_artifact_to_cache(cache, pkgs[artifact], channel, subdir, artifact)
        if i % 100 == 99:
            # save the state occasionally
            _save_cache(cache, cachefile, display=False)
    _save_cache(cache, cachefile)
    return cache


def generate_map(cache, channel, subdir, remove_exprs=DEFAULT_REMOVE_EXPRS, write=True):
    """Creates the map file from the cache.

    Parameters
    ----------
    cache : dict
        Cache contents, as described in readme.
    channel : str
        Name of the channel
    subdir : str
        Name of the platform subdir
    remove_exprs : sequence of str, optional
        Regular expressions for which command name matches are removed from the
        map file.
    write : bool, optional
        Flag for whether to actually write the file or not.
    """
    print(f"Generating map file for {channel}/{subdir}")
    # create remove expr func
    rm_res = list(map(re.compile, remove_exprs))

    def filter(exe):
        return not any([rm_re.match(exe) for rm_re in rm_res])

    # generate file
    exe_pkg = set()
    for pkg in cache.values():
        exe_pkg.update({(exe, pkg["package"]) for exe in pkg["executables"]})
    lines = sorted(exe_pkg)
    lines = [f"{e}:{p}" for e, p in lines if filter(e)]
    if lines:
        lines[-1] += "\n"
    s = "\n".join(lines)
    if write:
        channel_name = _get_channel_name(channel)
        mapfile = f"{channel_name}.{subdir}.map"
        print(f"Writing map file {mapfile}")
        with open(mapfile, "w") as f:
            f.write(s)
    return s


def generate(channel, remove_exprs=DEFAULT_REMOVE_EXPRS):
    """Generates the map files, and the cache files incidentally."""
    for subdir in SUBDIRS:
        cache = make_cache(channel, subdir)
        generate_map(cache, channel, subdir, remove_exprs=remove_exprs)



"""Main entry point for conda-suggest."""
import sys
import argparse


def make_parser():
    """Makes the argument parser"""
    p = argparse.ArgumentParser("conda-suggest")
    subcmd = p.add_subparsers(dest="subcmd", help="subcommand to execute")
    # Generate commands
    gen = subcmd.add_parser("generate", help="create map files for a channel")
    gen.add_argument("channel", help="Name of the channel")
    gen.add_argument("--remove-exprs", help="Regular expressions for which "
        "command name matches are removed from the map file", nargs="+",
        default=None, dest="remove_exprs", metavar="EXPR")
    # Find commands
    msg = subcmd.add_parser(
        "message",
        help="searches for the package an executable comes from and prints a "
        "'command-not-found' message.",
    )
    msg.add_argument("exe", help="Name of executable to find")
    return p


def main(args=None):
    """Main entry point function for conda-suggest."""
    p = make_parser()
    ns = p.parse_args(args=args)
    if ns.subcmd == "message":
        from . import find

        find.message(ns.exe)
    elif ns.subcmd == "generate":
        from . import generate

        if ns.remove_exprs is None:
            ns.remove_exprs = generate.DEFAULT_REMOVE_EXPRS
        generate.generate(ns.channel, remove_exprs=ns.remove_exprs)


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="conda-suggest",
    version="0.1.1",
    author="Anthony Scopatz",
    description="Conda Suggest",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/conda-incubator/conda-suggest",
    packages=['conda_suggest'],
    license="BSD",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "conda-suggest = conda_suggest.main:main",
        ],
    },
)
